class Validator(object):
    def __init__(self, is_valid, errors=None):
        if errors is not None:
            self._errors = errors
        else:
            self._errors = []
        self.is_valid = is_valid

    def __repr__(self):
        return '<{} ({})>'.format(self.__class__.__name__, self.is_valid)

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return self.is_valid

    @property
    def errors(self):
        return self._errors

    @property
    def error_summary(self):
        return [(x, 1) for x in self._errors]


class ValidationError(object):
    def __init__(self, msg, line=None, column=None, path=None):
        self.original_msg = msg
        self.line = line
        self.column = column
        self.path = path

        self.summary = msg
        self.details = ''

    def __repr__(self):
        max_length = 50
        txt = str(self)
        if len(txt) > max_length:
            txt = txt[:max_length-3] + '...'
        return '<{} ({})>'.format(self.__class__.__name__, txt)

    def __str__(self):
        return self.summary

    @property
    def url(self):
        return None

    @property
    def message(self):
        return str(self)



import webbrowser
try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from lxml import etree as ET

from ..standard.schema import get_schema
from ..standard.xsd_schema import XSDSchema
from ..utils.abstract import GenericSet
from ..utils.exceptions import SchemaError
from ..utils.querybuilder import XPathQueryBuilder


class Activity(object):
    """Class representing an IATI activity."""

    def __init__(self, etree, dataset=None, schema=None):
        self.etree = etree
        self.dataset = dataset
        self._schema = schema
        self.version = self.schema.version

    def __repr__(self):
        id_ = self.iati_identifier
        id_ = id_ if id_ else '[No identifier]'
        return '<{} ({})>'.format(self.__class__.__name__, id_)

    def show(self):
        """Open a new browser tab to the d-portal.org page
        for this dataset.
        """
        params = {'aid': self.iati_identifier}
        url = 'http://d-portal.org/q.html?{}'.format(urlencode(params))
        webbrowser.open_new_tab(url)

    @property
    def schema(self):
        if not self._schema:
            # TODO: Add a schema guesser,
            # based on the activity XML
            pass
        return self._schema

    @property
    def xml(self):
        """Return the raw XML of this activity, as a byte-string."""
        return bytes(ET.tostring(self.etree, pretty_print=True))

    @property
    def iati_identifier(self):
        """Return the iati-identifier for this activity,
        or ``None`` if it isn't provided.
        """
        id_ = self.schema.iati_identifier().run(self.etree)
        if id_:
            return id_[0].strip()
        return None

    def validate_iati(self):
        etree = ET.Element('iati-activities')
        etree.set('version', self.version)
        etree.append(self.etree)
        xsd_schema = XSDSchema('activity', self.version)
        return xsd_schema.validate(etree)

    @property
    def id(self):  # pylint: disable=invalid-name
        """Alias of ``iati_identifier``."""
        return self.iati_identifier

    @property
    def title(self):
        """Return a list of titles for this activity."""
        return self.schema.title().run(self.etree)

    @property
    def description(self):
        """Return a list of descriptions for this activity."""
        return self.schema.description().run(self.etree)

    @property
    def location(self):
        """Return a list of locations for this activity."""
        return self.schema.location().run(self.etree)

    @property
    def sector(self):
        """Return a list of sectors for this activity."""
        return self.schema.sector().run(self.etree)

    @property
    def humanitarian(self):
        """Return True if the humanitarian flag is set for this activity."""
        return self.schema.humanitarian().run(self.etree)

    @property
    def planned_start(self):
        """Return the planned start date for this activity,
        as a python ``date``.
        """
        date = self.schema.planned_start().run(self.etree)
        return date[0] if date else None

    @property
    def actual_start(self):
        """Return the actual start date for this activity,
        as a python ``date``.
        """
        date = self.schema.actual_start().run(self.etree)
        return date[0] if date else None

    @property
    def start(self):
        """Return the actual start date for this activity,
        if present. Otherwise, return the planned start.
        """
        start = self.actual_start
        if start:
            return start
        return self.planned_start

    @property
    def planned_end(self):
        """Return the planned end date for this activity,
        as a python ``date``.
        """
        date = self.schema.planned_end().run(self.etree)
        return date[0] if date else None

    @property
    def actual_end(self):
        """Return the actual end date for this activity,
        as a python ``date``.
        """
        date = self.schema.actual_end().run(self.etree)
        return date[0] if date else None

    @property
    def end(self):
        """Return the actual end date for this activity,
        if present. Otherwise, return the planned end.
        """
        end = self.actual_end
        if end:
            return end
        return self.planned_end


class ActivitySet(GenericSet):
    """Class representing a grouping of ``Activity`` objects.

    Objects in this grouping can be filtered and iterated over.
    Queries are only constructed and run when needed, so they
    can be efficient.
    """

    _key = 'iati_identifier'
    _multi_filters = [
        'id', 'iati_identifier', 'title', 'description',
        'location', 'sector', 'planned_start',
        'actual_start', 'planned_end', 'actual_end',
        'xpath', 'humanitarian',
    ]
    _instance_class = Activity
    _filetype = 'activity'
    _element = '/iati-activities/iati-activity'

    def __init__(self, datasets, **kwargs):
        super(ActivitySet, self).__init__()
        self.wheres = kwargs
        self.datasets = datasets

    def __len__(self):
        total = 0
        for dataset in self.datasets:
            if dataset.filetype != self._filetype:
                continue
            if not dataset.validate_xml():
                continue
            try:
                schema = get_schema(dataset.filetype, dataset.version)
            except SchemaError:
                continue
            prefix = self._element
            query = XPathQueryBuilder(
                schema,
                prefix=prefix,
                count=True
            ).where(**self.wheres)
            total += int(dataset.etree.xpath(query))
        return total

    def _query(self, schema=None):
        if schema is None:
            schema = get_schema(self._filetype, '2.03')
        return XPathQueryBuilder(
            schema,
            prefix=self._element,
        ).where(**self.wheres)

    def __iter__(self):
        for dataset in self.datasets:
            if dataset.filetype != self._filetype:
                continue
            if not dataset.validate_xml():
                continue
            try:
                schema = get_schema(dataset.filetype, dataset.version)
            except SchemaError:
                continue
            activity_etrees = dataset.etree.xpath(self._query(schema))
            for tree in activity_etrees:
                yield self._instance_class(tree, dataset, schema)



import json
from os.path import abspath, dirname, exists, join
import shutil
import tempfile
from unittest import TestCase

from mock import patch

from iatikit.utils import download
from iatikit.utils.config import CONFIG


XSD_TMPL = '''<?xml version="1.0" encoding="utf-8"?>

<xsd:schema xmlns:xsd="http://www.w3.org/2001/XMLSchema">

  <xsd:annotation>
    <xsd:documentation xml:lang="en">
        {filename}
    </xsd:documentation>
  </xsd:annotation>
'''


class MockRequest():
    def __init__(self, url):
        filename = url.rsplit('/', 1)[-1]
        if filename == 'Version.json':
            self.filepath = join(dirname(abspath(__file__)),
                                 'fixtures', 'codelist_downloads',
                                 'Version-v201.json')
        self.content = XSD_TMPL.format(filename=filename).encode()

    def json(self):
        with open(self.filepath) as handler:
            return json.load(handler)

    def raise_for_status(self):
        pass

class TestDownloadSchemas(TestCase):
    def setUp(self):
        self.standard_path = tempfile.mkdtemp(dir=dirname(abspath(__file__)))
        config_dict = {'paths': {'standard': self.standard_path}}
        CONFIG.read_dict(config_dict)

    @patch('requests.Session')
    def test_download_schemas(self, mock_session):
        mock_session.return_value.get.side_effect = MockRequest

        download.schemas()

        filenames = [
            'iati-activities-schema.xsd', 'iati-organisations-schema.xsd',
            'iati-common.xsd', 'xml.xsd',
        ]
        for version in ['201', '105', '104', '103', '102', '101']:
            for filename in filenames:
                filepath = join(self.standard_path, 'schemas',
                                version, filename)
                assert exists(filepath)
                with open(filepath) as handler:
                    contents = handler.read()
                assert contents == XSD_TMPL.format(filename=filename)

    def tearDown(self):
        shutil.rmtree(self.standard_path, ignore_errors=True)



from os.path import abspath, dirname, join
from unittest import TestCase

from mock import patch
from lxml import etree as ET

from iatikit.data.dataset import DatasetSet, Dataset
from iatikit.data.organisation import OrganisationSet, Organisation
from iatikit.standard.organisation_schema import OrganisationSchema203
from iatikit.utils.config import CONFIG


class TestOrganisationSet(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestOrganisationSet, self).__init__(*args, **kwargs)
        registry_path = join(dirname(abspath(__file__)),
                             'fixtures', 'registry')
        org_datasets = DatasetSet(
            join(registry_path, 'data', 'fixture-org', '*'),
            join(registry_path, 'metadata', 'fixture-org', '*'),
        )
        self.fixture_org_orgs = OrganisationSet(org_datasets)

        standard_path = join(dirname(abspath(__file__)), 'fixtures',
                             'standard')
        config_dict = {'paths': {'standard': standard_path}}
        CONFIG.read_dict(config_dict)

    def test_organisations_iter(self):
        organisations = list(self.fixture_org_orgs)
        assert len(organisations) == 1

    def test_organisations_len(self):
        assert len(self.fixture_org_orgs) == 1

    def test_organisations_filter_by_id(self):
        iati_id = 'GB-COH-01234567'
        orgs = self.fixture_org_orgs.where(id=iati_id).all()
        assert len(orgs) == 1
        assert orgs[0].org_identifier == iati_id

        not_iati_id = 'GB-COH-01234567-No-identifier'
        orgs = self.fixture_org_orgs.where(id=not_iati_id).all()
        assert len(orgs) == 0

    def test_organisations_filter_by_iati_identifier(self):
        iati_id = 'GB-COH-01234567'
        orgs = self.fixture_org_orgs.where(org_identifier=iati_id).all()
        assert len(orgs) == 1
        assert orgs[0].org_identifier == iati_id

        not_iati_id = 'GB-COH-01234567-No-identifier'

        orgs = self.fixture_org_orgs.where(org_identifier=not_iati_id).all()
        assert len(orgs) == 0


class TestOrganisation(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestOrganisation, self).__init__(*args, **kwargs)
        data_path = join(dirname(abspath(__file__)),
                         'fixtures', 'registry', 'data', 'fixture-org',
                         'fixture-org-org.xml')
        etree = ET.parse(data_path)
        org_el = etree.xpath('//iati-organisation')[0]
        dataset = Dataset(data_path, None)
        schema = OrganisationSchema203()
        self.organisation = Organisation(org_el, dataset, schema)

    @patch('webbrowser.open_new_tab')
    def test_activity_show(self, fake_open_new_tab):
        url = 'http://d-portal.org/ctrack.html' + \
              '?publisher=GB-COH-01234567#view=main'
        self.organisation.show()
        fake_open_new_tab.assert_called_once_with(url)

    def test_organisation_xml(self):
        xml = self.organisation.xml
        first_line = '<iati-organisation last-updated-datetime=' + \
                     '"2018-10-23T22:27:45" default-currency="GBP" ' + \
                     'xml:lang="en">'
        assert xml.decode().split('\n', 1)[0] == first_line

    def test_organisation_id(self):
        id_ = 'GB-COH-01234567'
        assert self.organisation.id == id_
        assert self.organisation.org_identifier == id_

    def test_organisation_repr(self):
        org_repr = '<Organisation (GB-COH-01234567)>'
        assert str(self.organisation) == org_repr



from os.path import abspath, dirname, join
from unittest import TestCase

from mock import patch
import pytest

from iatikit.data.publisher import PublisherSet, Publisher
from iatikit.utils.exceptions import FilterError


class TestPublishers(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestPublishers, self).__init__(*args, **kwargs)
        self.registry_path = join(dirname(abspath(__file__)),
                                  'fixtures', 'registry')

    def setUp(self):
        self.publishers = PublisherSet(
            join(self.registry_path, 'data', '*'),
            join(self.registry_path, 'metadata', '*'),
        )

    def test_publishers_iter(self):
        publisher_list = list(self.publishers)
        assert len(publisher_list) == 2

    def test_publishers_filter(self):
        bad_filtered_pubs = self.publishers.where(name='does-not-exist')
        empty_publisher_list = list(bad_filtered_pubs)
        assert empty_publisher_list == []

        filtered_pubs = self.publishers.where(name='fixture-org')
        publisher_list = list(filtered_pubs)
        assert len(publisher_list) == 1
        assert publisher_list[0].name == 'fixture-org'

    def test_publishers_unknown_filter(self):
        with pytest.raises(FilterError):
            self.publishers.find(unknown='unknown')

    def test_publishers_index_slicing(self):
        publisher_list = self.publishers[:]
        assert len(publisher_list) == 2

    def test_publishers_indexing(self):
        publisher_names = ['fixture-org', 'old-org']
        assert self.publishers[0].name in publisher_names
        assert self.publishers[1].name in publisher_names

    def test_publisher_from_publishers(self):
        fixture_org = self.publishers.get('fixture-org')
        assert fixture_org.name == 'fixture-org'
        assert fixture_org.datasets.count() == 3
        assert fixture_org.metadata.get(
            'publisher_iati_id') == 'GB-COH-01234567'

        old_org = self.publishers.get('old-org')
        assert old_org.name == 'old-org'
        assert old_org.datasets.count() == 2
        assert old_org.metadata.get(
            'publisher_iati_id') == 'NL-CHC-98765'


class TestPublisher(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestPublisher, self).__init__(*args, **kwargs)
        self.registry_path = join(dirname(abspath(__file__)),
                                  'fixtures', 'registry')

    def setUp(self):
        self.fixture_org = Publisher(
            join(self.registry_path, 'data', 'fixture-org'),
            join(self.registry_path, 'metadata', 'fixture-org'),
            join(self.registry_path, 'metadata', 'no-metadata.json'),
        )
        self.old_org = Publisher(
            join(self.registry_path, 'data', 'old-org'),
            join(self.registry_path, 'metadata', 'old-org'),
            join(self.registry_path, 'metadata', 'old-org.json'),
        )

    def test_publisher_name(self):
        assert self.fixture_org.name == 'fixture-org'

    def test_publisher_repr(self):
        publisher_repr = '<Publisher (fixture-org)>'
        assert str(self.fixture_org) == publisher_repr

    @patch('webbrowser.open_new_tab')
    def test_publisher_show(self, fake_open_new_tab):
        url = 'https://iatiregistry.org/publisher/old-org'
        self.old_org.show()
        fake_open_new_tab.assert_called_once_with(url)

    def test_datasets(self):
        old_org_datasets = self.old_org.datasets.all()
        assert len(old_org_datasets) == 2
        for dataset in old_org_datasets:
            assert dataset.name in [
                'old-org-acts',
                'old-org-missing-acts']

    def test_activities(self):
        assert self.old_org.activities.count() == 2
        old_org_activity = self.old_org.activities[0]
        assert old_org_activity.id == 'NL-CHC-98765-NL-CHC-98765-XX0D9001'

    def test_metadata(self):
        old_org_metadata = self.old_org.metadata
        assert old_org_metadata.get('publisher_country') == 'NL'

    def test_no_metadata(self):
        assert self.fixture_org.metadata == {}



from ..utils.exceptions import SchemaError
from ..utils.types import StringType, DateType, SectorType, XPathType, \
                          BooleanType
from ..utils.abstract import GenericType


# pylint: disable=too-many-ancestors
class ActivitySchema101(object):
    version = '1.01'

    @classmethod
    def xpath(cls):
        return XPathType('')

    @classmethod
    def id(cls):  # pylint: disable=invalid-name
        return cls.iati_identifier()

    @classmethod
    def iati_identifier(cls):
        return StringType('iati-identifier/text()')

    @classmethod
    def title(cls):
        return StringType('title/text()')

    @classmethod
    def description(cls):
        return StringType('description/text()')

    @classmethod
    def location(cls):
        return GenericType('location')

    @classmethod
    def sector(cls):
        # TODO: This should include lookups for other v1.0x vocabs
        condition = {
            '1': [None, 'DAC'],
            '2': 'DAC-3',
        }
        return SectorType('sector', condition)

    @classmethod
    def planned_start(cls):
        return DateType('activity-date[@type="start-planned"]/@iso-date')

    @classmethod
    def actual_start(cls):
        return DateType('activity-date[@type="start-actual"]/@iso-date')

    @classmethod
    def planned_end(cls):
        return DateType('activity-date[@type="end-planned"]/@iso-date')

    @classmethod
    def actual_end(cls):
        return DateType('activity-date[@type="end-actual"]/@iso-date')

    @classmethod
    def humanitarian(cls):
        return BooleanType('false')


class ActivitySchema102(ActivitySchema101):
    version = '1.02'


class ActivitySchema103(ActivitySchema102):
    version = '1.03'


class ActivitySchema104(ActivitySchema103):
    version = '1.04'


class ActivitySchema105(ActivitySchema104):
    version = '1.05'


class ActivitySchema201(ActivitySchema105):
    version = '2.01'

    @classmethod
    def title(cls):
        return StringType('title/narrative/text()')

    @classmethod
    def description(cls):
        return StringType('description/narrative/text()')

    @classmethod
    def sector(cls):
        condition = {
            '1': [None, '1'],
        }
        return SectorType('sector', condition)

    @classmethod
    def planned_start(cls):
        return DateType('activity-date[@type="1"]/@iso-date')

    @classmethod
    def actual_start(cls):
        return DateType('activity-date[@type="2"]/@iso-date')

    @classmethod
    def planned_end(cls):
        return DateType('activity-date[@type="3"]/@iso-date')

    @classmethod
    def actual_end(cls):
        return DateType('activity-date[@type="4"]/@iso-date')


class ActivitySchema202(ActivitySchema201):
    version = '2.02'

    @classmethod
    def humanitarian(cls):
        return BooleanType('@humanitarian')


class ActivitySchema203(ActivitySchema202):
    version = '2.03'


def get_activity_schema(version):
    schema = None

    if version == '2.03':
        schema = ActivitySchema203
    elif version == '2.02':
        schema = ActivitySchema202
    elif version == '2.01':
        schema = ActivitySchema201
    elif version == '1.05':
        schema = ActivitySchema105
    elif version == '1.04':
        schema = ActivitySchema104
    elif version == '1.03':
        schema = ActivitySchema103
    elif version == '1.02':
        schema = ActivitySchema102
    elif version == '1.01':
        schema = ActivitySchema101

    if not schema:
        msg = 'Unknown activity schema: version: {}'.format(version)
        raise SchemaError(msg)

    return schema



# from io import BytesIO
# import os
# from os.path import abspath, dirname, join
# import shutil
# import tempfile
# from unittest import TestCase
# import zipfile

# from mock import patch

# from iatikit.utils import download
# from iatikit.utils.config import CONFIG


# class MockRequest():
#     def __init__(self, url, stream=False):
#         registry_path = join(dirname(abspath(__file__)),
#                              'fixtures', 'registry')
#         self.raw = BytesIO()
#         with zipfile.ZipFile(self.raw, 'w') as ziph:
#             for root, _, files in os.walk(registry_path):
#                 for file in files:
#                     fullpath = join(root, file)
#                     ziph.write(fullpath, join('/', 'iati-data-main', fullpath[len(registry_path) + 1:]))
#         self.raw.seek(0)


# class TestDownloadData(TestCase):
#     def setUp(self):
#         self.data_path = tempfile.mkdtemp(dir=dirname(abspath(__file__)))
#         config_dict = {'paths': {'registry': self.data_path}}
#         CONFIG.read_dict(config_dict)

#     @patch('requests.get', MockRequest)
#     def test_download_data(self):
#         def list_files(path):
#             all_files = []
#             for root, _, files in os.walk(path):
#                 for file in files:
#                     all_files.append(join(root, file)[len(path):])
#             return all_files

#         download.data()

#         registry_path = join(dirname(abspath(__file__)),
#                              'fixtures', 'registry')
#         source_files = list_files(registry_path)
#         dest_files = list_files(self.data_path)

#         assert len(source_files) == len(dest_files)
#         for dest_file in dest_files:
#             assert dest_file in source_files

#     def tearDown(self):
#         shutil.rmtree(self.data_path, ignore_errors=True)



import json
from os.path import abspath, dirname, join
import re


class CodelistMockRequest():
    def __init__(self, url):
        codelist_path = join(dirname(abspath(__file__)),
                             'fixtures', 'codelist_downloads')

        match = re.search(r'(\d{3}).*?/([^\.\/]+)\.(csv|json)', url)
        if match:
            version, name, extension = match.groups()
            fname = '{name}-v{version}.{extension}'.format(
                name=name,
                version=version,
                extension=extension)
        else:
            raise NotImplementedError()
        self.filepath = join(codelist_path, fname)

    def json(self):
        with open(self.filepath) as handler:
            return json.load(handler)

    def iter_lines(self):
        with open(self.filepath, 'rb') as handler:
            return handler.readlines()

    def raise_for_status(self):
        pass


import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

http_adapter = HTTPAdapter(max_retries=Retry(total=3))


def get_iati_versions():
    session = requests.Session()
    session.mount('https://', http_adapter)
    versions_url = 'http://reference.iatistandard.org/201/codelists/' + \
                   'downloads/clv2/json/en/Version.json'
    versions_response = session.get(versions_url)
    versions_response.raise_for_status()
    versions = [
        d['code']
        for d in versions_response.json()['data']
    ]
    versions.reverse()
    return versions
